/*
   txt2esp.c - a program to convert a text file with reference
   commands into a valid esp file.
   Peter C. Capasso
*/


#include "hacks.h"
#ifndef PI
  #include <math.h>
#endif

#define MAX_STR 255


/* global vars */
int master_file_size = 79837557; /* this will change with each patch */
char master_file_name[] = "Morrowind.esm";
char seed_file_name[] = "seedhold.dat";


unsigned long for_rewrite;  /* index into output file header to put file size
                               info.  We'll need this for the fseek call at
                               the end */
unsigned long frmr_count_loc; /* fseek again */


/* forward references */
void usage();
void usage_exit();
unsigned char get_rnd_byte();
void set_seed();
void update_seed();
long rnd(long);
int write_intel_long(FILE *, int);
int write_intel_float(FILE *, float);
int read_line(FILE *, char *);
void trim_spaces(char *);
int get_quoted_string(char *, char *);
int parse_3vector(char *, int *);  /* integer vector! */
int parse_float(char *, float *);
int build_reference(char *, int, FILE *);
int write_nulls(FILE *, int);
int write_esp_header(FILE *, char *);
int build_tel_ref(char *, int, FILE *);


/* start of actual program */

void usage()
{
  fprintf(stderr, "usage: txt2esp4 <in_text_file> <out_esp_file> \
\"cell name\"\n");
}


void usage_exit()
{
  usage();
  exit(DOOPS);
}


unsigned char get_rnd_byte()
{
  unsigned int temp, dummy;

  dummy = rand();
  temp = rand();
  dummy = rand();
  dummy = rand();
  return( (temp mod 256) );
}


void set_seed()
{
  unsigned int time_value;  /* had better be 4 bytes */
  unsigned int x;
  unsigned char dummy;
  int status;
  FILE *seed_file;

  if ((seed_file = fopen(seed_file_name, "rb")) eq NULL) {
    /* seed file does not exist.  Just quietly do without it */
    x = 0;
  } else {  /* have seed file, so use it */
    status = fread(&x, sizeof(int), 1, seed_file);
    if (status eq 0) { /* read was not OK - do without it */
      fprintf(stderr, "warning: error reading seed hold file\n");
      x = 0;
    }
    fclose(seed_file);
  }
  time_value = time(NULL);    /* get time */
  time_value ^= x;       /* mix with stored random seed, if there is one */
  printf("mixed random seed is: %u\n", time_value);
  srand(time_value);
  x = 513;
  while (x--)  /* warm up rnd number generator */
    dummy = get_rnd_byte();
}


void update_seed()
{
  unsigned long new_seed;
  int status;
  FILE *seed_file;

  new_seed = (unsigned long) get_rnd_byte();
  new_seed *= 256;
  new_seed += (unsigned long) get_rnd_byte();
  new_seed *= 256;
  new_seed += (unsigned long) get_rnd_byte();
  new_seed *= 256;
  new_seed += (unsigned long) get_rnd_byte();

  seed_file = fopen(seed_file_name, "wb");
  if (status eq NULL) {
    fprintf(stderr, "warning: error creating seed hold file\n");
  } else {
    status = fwrite(&new_seed, sizeof(unsigned long), 1, seed_file);
    if (status neq 1) { /* write error */
      fprintf(stderr, "warning: error writing seed hold file\n");
      fclose(seed_file);
      remove(seed_file_name); /* delete the file if you can */
    } else { /* no write error */
      fclose(seed_file);
    }
  }
}


long rnd(r)
long r;
{
  unsigned long temp;
  unsigned char *ptr;

  ptr = (char *) &temp;
  /* a long is four bytes */
  *(ptr++) = get_rnd_byte();
  *(ptr++) = get_rnd_byte();
  *(ptr++) = get_rnd_byte();
  *ptr = get_rnd_byte();
  ptr = (char *) NULL;  /* discard pointer */
  temp = temp mod r;  /* not statistically pure, but adequate */
  return((long) temp);
}


int write_intel_long(f, n)
FILE *f;
int n;
{
  int status;

  status = fwrite( (char *) &n, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}


int write_intel_float(f, fL)
FILE *f;
float fL;
{
  int status;

  /* we got lucky.  in this compiler, it uses the same float format
     as the esm file.  so no translation is needed at all. */
  status = fwrite( (char *) &fL, 4, 1, f);
  if (status neq 1)
    return(DOOPS);
  return(OKDOKEY);
}


/* read_line
   will read a complete line, all the way to the end of line.
   if the line is longer than the string, the string will be truncated
   but the file pointer will be pointing to the first character of the
   next line */
int read_line(f, s)
FILE *f;
char *s;
{
  int i,c;

  for (i = 0;;i++) {
    if ((c = fgetc(f)) eq EOF) break;
    if (c eq '\n') break;
    if ((i - 1) < MAX_STR) {
      s[i] = c;
    }
  }
  if (i >= MAX_STR)
    i = (MAX_STR - 1);
  s[i] = 0;
  if ((c eq EOF) && (i eq 0))
    return(c);
  return(OKDOKEY);
}


void trim_spaces(s)
char *s;
{
  int lead_s, end_s, s_len;

  s_len = strlen(s);
  if (s_len) {
    /* first deal with leading space */
    lead_s = 0;
    while (lead_s < s_len) {
      if (s[lead_s] > ' ')
        break;
      lead_s++;
    }
    memmove(s, &s[lead_s], 1 + (s_len - lead_s)); /* will copy null if zero len str */
    /* leading space gone.  now see to the trailing spaces */
    end_s = strlen(s);
    if (end_s > 0) {
      while (end_s-- > 0) {
        if (s[end_s] > ' ') { /* if non space .... */
          s[++end_s] = '\0';  /* put null on end of string */
          end_s = 0;          /* signal for end of loop */
        }
      }
    }
  } /* end if s_len */
}


int get_quoted_string(s, obj_name)
char *s;
char *obj_name;
{
  char *ptr;
  int x;

  if (strlen(s) < 2) {
    fprintf(stderr, "get_quoted_string: string to short or < 2 quotes\n");
    return(DOOPS);
  }
  /* since leading space has been removed, first char must be a " */
  if (s[0] neq '"') {
    fprintf(stderr, "get_quoted_string: no leading quote\n");
    return(DOOPS);
  }
  /* first char is a quote.  Now find the ending quote, if it exists */
  ptr = strchr(&s[1], '"');
  if (ptr eq NULL) {
    fprintf(stderr, "get_quoted_string: no ending quote\n");
    return(DOOPS);
  }
  if ((ptr - s) < 2) { /* pointer to int, careful */
    fprintf(stderr, "get_quoted_string: null string (2 quotes togther)\n");
    return(DOOPS);
  }
  memmove( &obj_name[0], &s[1], ((ptr - s) - 1) ); /* copy name out */
  obj_name[(ptr - s) - 1] = '\0';  /* tack EOL onto name! */
  memmove( &s[0], &ptr[1], 1 + strlen(&ptr[1]) );   /* delete name from parse string */
  ptr = (char *) NULL;  /* prevent eager-beaver optimizers from trashing */
  return(OKDOKEY);
}


int parse_3vector(s, vec_3)
char *s;
int *vec_3;  /* pointer to array of three ints */
{
  char *ptr;

  /* vector is <number number number>
     note that numbers are separated by spaces only and that there
     are no spaces between the brackers and their adjacent numbers */
  /* vector is not (at this time) floating point */
  /* input string has no leading spaces, could have anything after the
     trailing bracket */
  if (s[0] neq '<') {
    fprintf(stderr, "parse_3vector: no '<'\n");
    return(DOOPS);
  }
  ptr = strchr(s, ' ');  /* find space after first number */
  if ((ptr - s) < 2) {
    fprintf(stderr, "parse_3vector: first number too short or bad\n");
    return(DOOPS);
  }
  ptr[0] = '\0';  /* put zero right after first number */
  vec_3[0] = atoi(&s[1]);
  ptr++;
  memmove( s, ptr, (strlen(ptr) + 1) );
  trim_spaces(s);

  /* get second number */
  ptr = strchr(s, ' ');  /* find space after second number */
  if ((ptr - s) < 1) {
    fprintf(stderr, "parse_3vector: second number too short or bad\n");
    return(DOOPS);
  }
  ptr[0] = '\0';  /* put zero right after second number */
  vec_3[1] = atoi(s);
  ptr++;
  memmove( s, ptr, (strlen(ptr) + 1) );
  trim_spaces(s);

  /* get third number */
  ptr = strchr(s, '>');  /* find bracket after third number */
  if ((ptr - s) < 1) {
    fprintf(stderr, "parse_3vector: third number too short or bad\n");
    return(DOOPS);
  }
  ptr[0] = '\0';  /* put zero right after second number (replaces '>') */
  vec_3[2] = atoi(s);
  ptr++;
  memmove( s, ptr, (strlen(ptr) + 1) );

  return(OKDOKEY);
}


int parse_float(s, f)
char *s;
float *f;
{
  char *ptr;
  char num_temp[256];

  /* no leading spaces */
  /* look for space after the number */
  ptr = strchr(s, ' ');  /* find space after number */
  if (ptr eq NULL) {  /* entire string is number */
    *f = atof(s);
    return(DOOPS);  /* should be OKDOKEY - change *later* */
  }
  /* ptr points to first space after the number */
  memmove(&num_temp[0], s, (ptr - s));
  num_temp[((ptr - s) + 1)] = '\0';
  *f = atof(num_temp);
  memmove(s, ptr, strlen(ptr) + 1);
  return(OKDOKEY);
}


int build_reference(line_to_parse, object_num, f)
char *line_to_parse;
int object_num;
FILE *f;
{
  char object_name[256];
  float o_scale;     /* scaling factor (0.5 - 2.0) */
  int rot_vec[3];    /* rotation vector - degrees*/
  int trans_vec[3];  /* translation vector */
  int status;

  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = get_quoted_string(line_to_parse, &object_name[0]);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_float(line_to_parse, &o_scale);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_3vector(line_to_parse, (int *) &rot_vec[0]);  /* rotation */
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_3vector(line_to_parse, (int *) &trans_vec[0]);  /* translation (placement) */
  if (status eq DOOPS) return(status);

  /* Now write this chunk of the esp !EEP! */

  fprintf(f, "FRMR");
  status = write_intel_long(f, 4);
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, object_num); /* first object is 1, not 0 */
  if (status eq DOOPS) return(status);
  fprintf(f, "NAME");
  status = write_intel_long(f, strlen(object_name) + 1);
  if (status eq DOOPS) return(status);
  fprintf(f, object_name);
  status = fputc(0, f);   /* null at end of string */
  if (status eq EOF) return(status);
  if (o_scale neq 1.0) {
    fprintf(f, "XSCL");
    status = write_intel_long(f, 4);
    if (status eq DOOPS) return(status);
    status = write_intel_float(f, o_scale); /* scale, 4 byte float */
    if (status eq DOOPS) return(status);
  }
  fprintf(f, "DATA");
  status = write_intel_long(f, 24);
  if (status eq DOOPS) return(status);
  /* X, Y, Z translations */
  status = write_intel_float(f, (float) trans_vec[0]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_vec[1]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_vec[2]);
  if (status eq DOOPS) return(status);
  /* X, Y, Z rotations - SHIT - they GOTTA be in radians!*/
  status = write_intel_float(f, ((float) rot_vec[0] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_vec[1] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_vec[2] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  return(OKDOKEY);  
}


int write_nulls(f, how_many)
FILE *f;
int how_many;
{
  int status;

  while(how_many--)
    if ((status = fputc(0, f)) eq EOF)
      return(DOOPS);
}


int write_esp_header(f, cell_name)
FILE *f;
char *cell_name;
{
  int status, temp4, temp5;
  float ftemp;

  fprintf(f, "TES3");
  status = write_intel_long(f, 332 + strlen(master_file_name) + 1);
  if (status eq DOOPS) return(status);
  /* write 8 nulls */
  status = write_nulls(f, 8);
  if (status eq DOOPS) return(status);
  fprintf(f, "HEDR");
  status = write_intel_long(f, 300);
  if (status eq DOOPS) return(status);
  /* version number of esm file is a four byte float.
     This is set to the 1.2 (patched) version */
  status = write_intel_float(f, 1.2);
  if (status eq DOOPS) return(status);
  status = write_nulls(f, 4);
  if (status eq DOOPS) return(status);
  /* author name, 32 chars, all nulls for now */
  status = write_nulls(f, 32);  
  if (status eq DOOPS) return(status);
  /* description of file, 260 chars, all nulls for now */
  status = write_nulls(f, 260);  
  if (status eq DOOPS) return(status);
  fprintf(f, "MAST");
  /* length of master filename including null */
  status = write_intel_long(f, strlen(master_file_name) + 1);
  if (status eq DOOPS) return(status);
  fprintf(f, master_file_name);
  status = fputc(0, f); /* write null terminator */
  if (status eq EOF) return(status);
  fprintf(f, "DATA");
  status = write_intel_long(f, 8);
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, master_file_size); /* low four bytes */
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, 0);                /* high four bytes */
  if (status eq DOOPS) return(status);

  /* technically that is the end of the ESM file header.  Since for this
     program we are going to write one interior cell, no more no less,
     we are going to conceptually think of the one cell header as part of
     the esm file header even though this is incorrect.  So now we start
     on the cell header */

  fprintf(f, "CELL");
  for_rewrite = ftell(f); /* for later */
  status = write_nulls(f, 12); /* 12 byte cell size, to fill in later */
  if (status eq DOOPS) return(status);
  fprintf(f, "NAME");
  status = write_intel_long(f, strlen(cell_name) + 1); /* length of cell name including null at end */
  if (status eq DOOPS) return(status);
  fprintf(f, cell_name);   /* cell name */
  status = fputc(0, f); /* write the null terminator */
  if (status eq EOF) return(status);
  fprintf(f, "DATA");
  status = write_intel_long(f, 12); /* length of cell data */
  if (status eq DOOPS) return(status);

  status = fputc(1, f); /* first 4 bytes are a bitmask */
  if (status eq EOF) return(status);
  status = write_nulls(f, 3);
  if (status eq DOOPS) return(status);

  status = write_intel_long(f, 0);  /* "level of fog" (fog color?) */
  if (status eq DOOPS) return(status);

  ftemp = 0.8 + (0.1 * rnd(21));  /* random fog density */
  status = write_intel_float(f, ftemp);  /* fog density (0.0 - 1.0) */
  if (status eq DOOPS) return(status);

  fprintf(f, "INTV");
  status = write_intel_long(f, 4);
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, 0); /* water height, if water enabled */
  if (status eq DOOPS) return(status);

  fprintf(f, "AMBI");
  status = write_intel_long(f, 16); /* size of ambi chunk */
  if (status eq DOOPS) return(status);

  /* rgb ambient lighting */
  status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
  status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
  status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
  status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);

  /* rgb sunlight lighting */
  if (rnd(4) > 1) {
    temp4 = 18 + rnd(10);
    temp5 = 24 + rnd(25);
    status = fputc((temp4 + temp5), f);  if (status eq EOF) return(status);
    status = fputc(temp5, f);  if (status eq EOF) return(status);
    status = fputc(temp5, f);  if (status eq EOF) return(status);
  } else {
    status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
    status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
    status = fputc((25 + rnd(45)), f);  if (status eq EOF) return(status);
  }
  status = fputc(0, f);  if (status eq EOF) return(status);

  /* rgb fog color */
  status = fputc(rnd(35), f);  if (status eq EOF) return(status);
  status = fputc(rnd(35), f);  if (status eq EOF) return(status);
  status = fputc(rnd(35), f);  if (status eq EOF) return(status);
  status = fputc(0, f);  if (status eq EOF) return(status);

  status = write_intel_float(f, ftemp);  /* fog density (0.0 - 1.0) */
  if (status eq EOF) return(status);

  return(OKDOKEY);
}


int build_tel_ref(line_to_parse, object_num, f)
char *line_to_parse;
int object_num;
FILE *f;
{
  char object_name[256];
  char dest_name[256];
  float o_scale;      /* scaling factor (0.5 - 2.0) */
  int rot_vec[3];     /* rotation vector - degrees*/
  int trans_vec[3];   /* translation vector */
  int rot_markr[3];   /* rotation vector for destination marker */
  int trans_markr[3]; /* location of destination marker */
  int status;
  int is_interior;

  is_interior = 0;
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = get_quoted_string(line_to_parse, &object_name[0]);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_float(line_to_parse, &o_scale);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_3vector(line_to_parse, (int *) &rot_vec[0]);  /* rotation */
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_3vector(line_to_parse, (int *) &trans_vec[0]);  /* translation (placement) */
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  /* now for teleport data */
  status = parse_3vector(line_to_parse, (int *) &rot_markr[0]);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  status = parse_3vector(line_to_parse, (int *) &trans_markr[0]);
  if (status eq DOOPS) return(status);
  trim_spaces(line_to_parse);  /* trim leading and trailing spaces, again */

  /* at this point, if it is a teleport to an exterior cell, then we
     are done.  If it is a teleport to an interior cell, then the name
     in quotes will follow.  So if the line is empty, it is a valid
     exterior door.  If there is a quoted string, then we have a
     valid interior door.  If there is anything else, we have an error. */
  if (strlen(line_to_parse)) {
    is_interior = 1;
    status = get_quoted_string(line_to_parse, &dest_name[0]);
    if (status eq DOOPS) return(status);
  }

  /* Now write this chunk of the esp !EEP! */
  fprintf(f, "FRMR");
  status = write_intel_long(f, 4);
  if (status eq DOOPS) return(status);
  status = write_intel_long(f, object_num); /* first object is 1, not 0 */
  if (status eq DOOPS) return(status);
  fprintf(f, "NAME");
  status = write_intel_long(f, strlen(object_name) + 1);
  if (status eq DOOPS) return(status);
  fprintf(f, object_name);
  status = fputc(0, f);   /* null at end of string */
  if (status eq EOF) return(status);
  /* scale precedes DODT chunk */
  if (o_scale neq 1.0) {
    fprintf(f, "XSCL");
    status = write_intel_long(f, 4);
    if (status eq DOOPS) return(status);
    status = write_intel_float(f, o_scale); /* scale, 4 byte float */
    if (status eq DOOPS) return(status);
  }
  /* teleport information ... */
  fprintf(f, "DODT");
  status = write_intel_long(f, 24);
  if (status eq DOOPS) return(status);
  /* X, Y, Z translations */
  status = write_intel_float(f, (float) trans_markr[0]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_markr[1]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_markr[2]);
  if (status eq DOOPS) return(status);
  /* X, Y, Z rotations - SHIT - they GOTTA be in radians!*/
  status = write_intel_float(f, ((float) rot_markr[0] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_markr[1] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_markr[2] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  /* additonal info only for interior teleporting door */
  if (is_interior) {
    fprintf(f, "DNAM"); /* Destination cell NAMe */
    status = write_intel_long(f, strlen(dest_name) + 1);
    if (status eq DOOPS) return(status);
    fprintf(f, "%s", dest_name);
    status = fputc(0, f);   /* null at end of string */
    if (status eq EOF) return(status);    
  }
  fprintf(f, "DATA");
  status = write_intel_long(f, 24);
  if (status eq DOOPS) return(status);
  /* X, Y, Z translations */
  status = write_intel_float(f, (float) trans_vec[0]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_vec[1]);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, (float) trans_vec[2]);
  if (status eq DOOPS) return(status);
  /* X, Y, Z rotations - SHIT - they GOTTA be in radians!*/
  status = write_intel_float(f, ((float) rot_vec[0] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_vec[1] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  status = write_intel_float(f, ((float) rot_vec[2] * PI) / 180.0);
  if (status eq DOOPS) return(status);
  return(OKDOKEY);
}


main(argc, argv)
int argc;
char *argv[];
{
  FILE *f1, *f2;
  char line_to_parse[256];
  int object_count = 0;     /* each reference in the ESP has an index */
  int object_subtract = 0;  /* some refs are not in the count in the NAM0
                               chunk */
  int status;
  int parse_found;
  unsigned long where_you_at_now;

  if (argc neq 4)
    usage_exit();
  if ((f1 = fopen(argv[1], "r")) eq NULL) {
    fprintf(stderr, "Error opening input file \"%s\"\n", argv[1]);
    exit(DOOPS);
  }
  if ((f2 = fopen(argv[2], "wb")) eq NULL) {
    fprintf(stderr, "Error creating esp file \"%s\"\n", argv[2]);
    exit(DOOPS);
  }
  set_seed();  /* simple way to initialize random seed */
  write_esp_header(f2, argv[3]);  /* argv[3] has the cell name */

  /* right, ready to go ... */
  /* The eps format likes references to be ordered by type.  It wants
     references to NPCs first, then references to Creatures (but not
     leveled creatures), and then references to teliporting doors.
     After that comes the NAM0 chunk.  Following this is all of the
     "regular" references (in no particular oreder!), which include
     references to:
       - non teleporting doors
       - leveled creatures
       - containers
       - statics
       - lights
       - activators

     There is more to the ESP file but since this program will deal
     only with references to items in an interior cell, we ignore the
     rest.

     Anyway, because of the ordering, we make four passes over the
     input text file.  So the input file can't be a pipe or a stream.
   */

  /* loop 1 - for NPC references */
  while ((status = read_line(f1, &line_to_parse[0])) neq EOF) {
    trim_spaces(&line_to_parse[0]);  /* trim leading and trailing spaces */
    if (strlen(line_to_parse) > 0) { /* if line is blank, ignore and go to
                                        next line */
      parse_found = 0;
      if (0 eq strncmp("npc_ref ", line_to_parse, 8)) {
        parse_found = 1;
        /* we know it is a npc reference object.  Go to the npc reference
           object handler */
        /* toss the "npc_ref" token and the one known space after it */
        memmove(
          line_to_parse,
          &line_to_parse[8],
          (1 + strlen(&line_to_parse[8]))
        );
        /* right. the structure of an NPC reference is identical to
           a "regular" reference.  So long as it is before all the
           other types of references, all is fine.  Since we've tossed
           the "npc_ref" token, we can re-use the subroutine that
           handles "regular" references.  */
        status = build_reference(line_to_parse, ++object_count, f2);
        if (status eq DOOPS) {
          fprintf(stderr, "build_reference returned an error\n");
          exit(DOOPS);
        }
        object_subtract++;
      }
    }
  }
  status = fseek(f1, 0, SEEK_SET); /* try to rewind input file */
  if (status) {
    fprintf(stderr, "fseek error on input file\n");
    exit(DOOPS);
  }

  /* loop 2 - for references to creatures (but not leveled creatures) */
  while ((status = read_line(f1, &line_to_parse[0])) neq EOF) {
    trim_spaces(&line_to_parse[0]);  /* trim leading and trailing spaces */
    if (strlen(line_to_parse) > 0) { /* if line is blank, ignore and go to next line */
      parse_found = 0;
      if (0 eq strncmp("creature_ref ", line_to_parse, 13)) {
        parse_found = 1;
        /* we know it is a referenced creature.  Go to the reference object
           handler */
        /* toss the "creature_ref" token and the one known space after it */
        memmove(
          line_to_parse,
          &line_to_parse[13],
          (1 + strlen(&line_to_parse[13]))
        );
        /* again, subroutine re-use because structure of this reference
           is the same as a "regular" reference */
        status = build_reference(line_to_parse, ++object_count, f2); /* first index is 1 */
        if (status eq DOOPS) {
          fprintf(stderr, "build_reference returned an error\n");
          exit(DOOPS);
        }
        object_subtract++;
      }
    }
  }
  status = fseek(f1, 0, SEEK_SET); /* try to rewind input file */
  if (status) {
    fprintf(stderr, "fseek error on input file\n");
    exit(DOOPS);
  }

  /* loop 3 - for teliporting door references */
  while ((status = read_line(f1, &line_to_parse[0])) neq EOF) {
    trim_spaces(&line_to_parse[0]);  /* trim leading and trailing spaces */
    if (strlen(line_to_parse) > 0) { /* if line is blank, ignore and go to next line */
      parse_found = 0;
      if (0 eq strncmp("tel_door_ref ", line_to_parse, 13)) {
        parse_found = 1;
        /* we know it is a teleport door object.  Go to the teli door object
           handler */
        /* toss the "tel_door_ref" token and the one known space after it */
        memmove(
          line_to_parse,
          &line_to_parse[13],
          (1 + strlen(&line_to_parse[13]))
        );
        status = build_tel_ref(line_to_parse, ++object_count, f2); /* first index is 1 */
        if (status eq DOOPS) {
          fprintf(stderr, "build_tel_ref returned an error\n");
          exit(DOOPS);
        }
        object_subtract++;
      }
    }
  }
  status = fseek(f1, 0, SEEK_SET); /* try to rewind input file */
  if (status) {
    fprintf(stderr, "fseek error on input file\n");
    exit(DOOPS);
  }

  /* three passes have been done.  Now we write the NAM0 chunk, before
     doing the fourth pass */
  fprintf(f2, "NAM0");  /* might be the number of FRMR records in the file */
  status = write_intel_long(f2, 4); /* length of NAM0 chunk */
  if (status eq DOOPS) {
    fprintf(stderr, "error 1 writing NAM0 chunk\n");
    exit(DOOPS);
  }
  frmr_count_loc = ftell(f2);
  status = write_intel_long(f2, 0); /* fill this in later */
  if (status eq DOOPS) {
    fprintf(stderr, "error 2 writing NAM0 chunk\n");
    exit(DOOPS);
  }

  /* loop 4 - for "regular" references */
  while ((status = read_line(f1, &line_to_parse[0])) neq EOF) {
    trim_spaces(&line_to_parse[0]);  /* trim leading and trailing spaces */
    if (strlen(line_to_parse) > 0) { /* if line is blank, ignore and go to next line */
      parse_found = 0;
      if (0 eq strncmp("reference ", line_to_parse, 10)) {
        parse_found = 1;
        /* we know it is a reference object.  Go to the reference object
           handler */
        /* toss the "reference" token and the one known space after it */
        memmove(
          line_to_parse,
          &line_to_parse[10],
          (1 + strlen(&line_to_parse[10]))
        );
        status = build_reference(line_to_parse, ++object_count, f2); /* first index is 1 */
        if (status eq DOOPS) {
          fprintf(stderr, "build_reference returned an error\n");
          exit(DOOPS);
        }
      }
    }
  }

  /* At this point, we're so close to done it isn't funny, however,
     we have to write the file size to the header, so do an ftell,
     then rewind back a bit, and do some math, and write that into
     the cell header.  Also the NAM0 chunk needs to be updated. */
  where_you_at_now = ftell(f2);
  status = fseek(f2, for_rewrite, SEEK_SET);
  if (status) {
    fprintf(stderr, "fseek error at very end of processing.  FUCK!\n");
    exit(DOOPS);
  }
  status = write_intel_long(f2,  ((where_you_at_now - for_rewrite) - 12) );
  if (status eq DOOPS) {
    fprintf(stderr, "write error at very end of processing.  FUCK!\n");
    exit(DOOPS);
  }
  /* the cell length is actually 12 bytes, so we've we-written the first 4
     bytes and we'll leave the following 8 bytes as is (we started them as
     nulls/zeros, which now works out real well) */

  /* now update the count of FRMRs in the NAM0 chunk */
  status = fseek(f2, frmr_count_loc, SEEK_SET);
  if (status) {
    fprintf(stderr, "fseek error 2 at very end of process\n");
    exit(DOOPS);
  }
  status = write_intel_long(f2, (object_count - object_subtract));
  if (status eq DOOPS) {
    fprintf(stderr, "FUCK - write error 2 at very end of process!\n");
    exit(DOOPS);
  }

  /* NOW we are done */
  fclose(f2);     /* close destination file is HIGHESR priority */
  fclose(f1);     /* important but lower priority */
  update_seed();  /* lower */
  exit(OKDOKEY);  /* really really done now */
}


/* actual end of this file */
